"""Tests for the paragraph module."""

import docx
import pytest
from docx.text import paragraph as docx_paragraph

from cmi_docx import paragraph, styles


@pytest.fixture
def sample_paragraph() -> docx_paragraph.Paragraph:
    """Returns a sample paragraph."""
    document = docx.Document()
    return document.add_paragraph("This is a sample paragraph.")


def test_find_single_in_paragraph(sample_paragraph: docx_paragraph.Paragraph) -> None:
    """Test finding a text in a paragraph."""
    extend_paragraph = paragraph.ExtendParagraph(sample_paragraph)
    expected = paragraph.FindParagraph(
        paragraph=sample_paragraph,
        character_indices=[(10, 16)],
    )

    actual = extend_paragraph.find_in_paragraph("sample")

    assert actual == expected


def test_find_multiple_in_paragraph(sample_paragraph: docx_paragraph.Paragraph) -> None:
    """Test finding multiple texts in a paragraph."""
    extend_paragraph = paragraph.ExtendParagraph(sample_paragraph)
    expected = paragraph.FindParagraph(
        paragraph=sample_paragraph,
        character_indices=[(2, 4), (5, 7)],
    )

    actual = extend_paragraph.find_in_paragraph("is")

    assert actual == expected


def test_find_in_single_run(sample_paragraph: docx_paragraph.Paragraph) -> None:
    """Test finding a text in a single paragraph run."""
    extend_paragraph = paragraph.ExtendParagraph(sample_paragraph)
    expected = (
        paragraph.run.FindRun(
            paragraph=sample_paragraph,
            run_indices=(0, 0),
            character_indices=(2, 4),
        ),
        paragraph.run.FindRun(
            paragraph=sample_paragraph,
            run_indices=(0, 0),
            character_indices=(5, 7),
        ),
    )

    actual = extend_paragraph.find_in_runs("is")

    assert actual[0].paragraph.text == expected[0].paragraph.text
    assert actual[0].run_indices == expected[0].run_indices
    assert actual[0].character_indices == expected[0].character_indices
    assert actual[1].paragraph.text == expected[1].paragraph.text
    assert actual[1].run_indices == expected[1].run_indices
    assert actual[1].character_indices == expected[1].character_indices


def test_find_in_single_run_complete() -> None:
    """Tests finding an exact match of a run."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    para.add_run("full-run")
    para.add_run("another")
    extend_paragraph = paragraph.ExtendParagraph(para)

    actual = extend_paragraph.find_in_runs("full-run")

    assert len(actual[0].runs) == 1


def test_replace_single_run(sample_paragraph: docx_paragraph.Paragraph) -> None:
    """Test replacing text in a paragraph."""
    extend_paragraph = paragraph.ExtendParagraph(sample_paragraph)

    extend_paragraph.replace("sample", "example")

    assert sample_paragraph.text == "This is a example paragraph."


def test_replace_multiple_runs(
    sample_paragraph: docx_paragraph.Paragraph,
) -> None:
    """Test replacing text in multiple runs."""
    sample_paragraph.add_run(" This is a sample paragraph.")
    extend_paragraph = paragraph.ExtendParagraph(sample_paragraph)

    extend_paragraph.replace("This is", "That was")

    assert (
        sample_paragraph.text
        == "That was a sample paragraph. That was a sample paragraph."
    )


def test_insert_run_middle() -> None:
    """Test inserting a run into a paragraph."""
    document = docx.Document()
    para = document.add_paragraph("")
    para.add_run("Hello ")
    para.add_run("world!")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.insert_run(1, "beautiful ", styles.RunStyle(bold=True))

    assert para.text == "Hello beautiful world!"
    assert para.runs[1].bold


def test_insert_run_start() -> None:
    """Test inserting a run at the start of a paragraph."""
    document = docx.Document()
    para = document.add_paragraph("world!")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.insert_run(0, "Hello ", styles.RunStyle(bold=True))

    assert para.text == "Hello world!"
    assert para.runs[0].bold


@pytest.mark.parametrize("index", [-1, 1])
def test_insert_run_end(index: int) -> None:
    """Test inserting a run at the end of a paragraph."""
    document = docx.Document()
    para = document.add_paragraph("Hello")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.insert_run(index, " world!", styles.RunStyle(bold=True))

    assert para.text == "Hello world!"
    assert para.runs[1].bold


def test_insert_run_empty() -> None:
    """Test inserting a run into an empty paragraph."""
    document = docx.Document()
    para = document.add_paragraph("")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.insert_run(0, "Hello", styles.RunStyle(bold=True))

    assert para.text == "Hello"
    assert para.runs[0].bold


def test_replace_between_one_run() -> None:
    """Test replacing text in one run."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.replace_between(5, 7, "was")

    assert para.text == "This was a sample paragraph."
    assert para.runs[0].text == "This "
    assert para.runs[1].text == "was"
    assert para.runs[2].text == " a sample paragraph."


def test_replace_between_multiple_runs() -> None:
    """Test replacing text in multiple runs."""
    document = docx.Document()
    para = document.add_paragraph("This ")
    para.add_run("is")
    para.add_run(" Sparta!")
    extend_paragraph = paragraph.ExtendParagraph(para)

    extend_paragraph.replace_between(3, 10, "nk sm")

    assert para.text == "Think smarta!"
    assert para.runs[0].text == "Thi"
    assert para.runs[1].text == "nk sm"
    assert para.runs[2].text == "arta!"



"""Image components for declarative documents."""

import dataclasses
import pathlib
from collections.abc import Coroutine
from typing import Literal

from cmi_docx.declarative import base


@dataclasses.dataclass
class ImageRun(base.Component):
    """An image embedded in the document.

    Attributes:
        data: Image data as bytes, file path, or a coroutine that resolves to bytes.
        type: Image type (e.g., 'png', 'jpg', 'jpeg', 'bmp', 'gif', 'svg').
        transformation: Dictionary with 'width' and/or 'height' in points.
        alt_text: Alternative text for accessibility.
    """

    data: bytes | str | pathlib.Path | Coroutine[None, None, bytes]
    type: str | None = None
    transformation: dict[Literal["width", "height"], int | float] | None = None
    alt_text: dict[Literal["title", "description", "name"], str] | None = None



"""Basic tests for the declarative API."""

import pytest

from cmi_docx import declarative


@pytest.mark.asyncio
async def test_simple_document() -> None:
    """Test creating a simple document with text."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Hello World"),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.paragraphs[0].text.startswith("Hello World")


@pytest.mark.asyncio
async def test_document_with_text_runs() -> None:
    """Test creating a document with formatted text runs."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(
                        children=[
                            declarative.TextRun(text="Bold text", bold=True),
                            declarative.TextRun(text=" and "),
                            declarative.TextRun(text="italic text", italic=True),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.paragraphs[0].text.startswith("Bold text")
    assert docx.paragraphs[0].runs[0].bold
    assert docx.paragraphs[0].runs[2].italic


@pytest.mark.asyncio
async def test_document_with_heading() -> None:
    """Test creating a document with headings."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Main Heading", heading=1),
                    declarative.Paragraph(text="Subheading", heading=2),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.paragraphs[0].text.startswith("Main Heading")
    assert docx.paragraphs[0].style is not None
    assert docx.paragraphs[0].style.name == "Heading 1"
    assert docx.paragraphs[1].text.startswith("Subheading")
    assert docx.paragraphs[1].style is not None
    assert docx.paragraphs[1].style.name == "Heading 2"


@pytest.mark.asyncio
async def test_document_metadata() -> None:
    """Test creating a document with metadata."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Content"),
                ],
            ),
        ],
        title="Test Document",
        creator="Test Author",
        subject="Testing",
    )

    docx_doc = await doc.to_docx()
    assert docx_doc.core_properties.title == "Test Document"
    assert docx_doc.core_properties.author == "Test Author"
    assert docx_doc.core_properties.subject == "Testing"



""".. include:: ../../README.md"""  # noqa: D415

# sergey: disable-file: IMP001 # Allow importing non-modules for barrel export.

from cmi_docx import declarative  # noqa: F401
from cmi_docx.comment import add_comment  # noqa: F401
from cmi_docx.document import ExtendDocument  # noqa: F401
from cmi_docx.paragraph import ExtendParagraph, FindParagraph  # noqa: F401
from cmi_docx.run import ExtendRun, FindRun  # noqa: F401
from cmi_docx.styles import (  # noqa: F401
    CellBorder,
    CellStyle,
    ParagraphStyle,
    RunStyle,
    TableSections,
    TableStyle,
)
from cmi_docx.table import ExtendCell, ExtendTable  # noqa: F401



"""Table components for declarative documents."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from cmi_docx.declarative import base, paragraph

if TYPE_CHECKING:
    from collections.abc import Coroutine, Iterable


@dataclasses.dataclass
class TableCell(base.Component):
    """A table cell containing paragraphs or nested tables.

    Attributes:
        children: List of Paragraph or Table components, or coroutines that
            resolve to these types.
        borders: Cell border configuration.
    """

    children: (
        Iterable[
            paragraph.Paragraph
            | Table
            | Coroutine[None, None, paragraph.Paragraph | Table]
        ]
        | None
    ) = None
    borders: dict[str, dict[str, str | int]] | None = None


@dataclasses.dataclass
class TableRow(base.Component):
    """A table row containing cells.

    Attributes:
        children: List of TableCell components or coroutines that resolve to cells.
    """

    children: list[TableCell | Coroutine[None, None, TableCell]]


@dataclasses.dataclass
class Table(base.Component):
    """A table with rows and cells.

    Attributes:
        rows: List of TableRow components or coroutines that resolve to rows.
        column_widths: List of column widths in DXA units.
        width: Table width configuration (dict with 'size' and 'type' keys).
        borders: Table border configuration.
        alignment: Table alignment ('left', 'center', 'right').
        indent: Table indent from left margin.
        layout: Table layout type ('autofit', 'fixed').
        style: Table style name.
    """

    rows: list[TableRow | Coroutine[None, None, TableRow]]
    column_widths: list[int] | None = None
    width: dict[str, int | str] | None = None
    borders: dict[str, dict[str, str | int]] | None = None
    alignment: str | None = None
    indent: int | None = None
    layout: str | None = None
    style: str | None = None



"""Section, header, and footer components for declarative documents."""

import dataclasses
from collections.abc import Coroutine
from typing import Literal

from cmi_docx.declarative import base, paragraph, table

type BlockElement = paragraph.Paragraph | table.Table
type HeaderFooterType = Literal["default", "first", "even"]


@dataclasses.dataclass
class SectionProperties:
    """Configuration for a document section.

    Attributes:
        page_size: Page dimensions.
        page_orientation: 'portrait' or 'landscape'.
        page_numbering: Page numbering configuration.
        columns: Column configuration (dict with 'count', 'space', 'separator').
        vertical_align: Vertical alignment ('top', 'center', 'bottom', 'both').
        title_page: Use different header/footer on first page.
        type: Section break type ('nextPage', 'nextColumn', 'continuous',
            'evenPage', 'oddPage').
    """

    page_size: dict[Literal["width", "height"], int] | None = None
    page_orientation: Literal["portrait", "landscape"] | None = None
    page_numbering: dict[Literal["start", "format"], int | str] | None = None
    columns: dict[Literal["count", "space", "separator"], int | bool] | None = None
    vertical_align: Literal["top", "center", "bottom", "both"] | None = None
    title_page: bool | None = None
    type: (
        Literal["nextPage", "nextColumn", "continuous", "evenPage", "oddPage"] | None
    ) = None


@dataclasses.dataclass
class Header(base.Component):
    """A section header.

    Attributes:
        children: List of Paragraph or Table components, or coroutines that
            resolve to these types.
    """

    children: (
        list[
            paragraph.Paragraph
            | table.Table
            | Coroutine[None, None, paragraph.Paragraph | table.Table]
        ]
        | None
    ) = None


@dataclasses.dataclass
class Footer(base.Component):
    """A section footer.

    Attributes:
        children: List of Paragraph or Table components, or coroutines that
            resolve to these types.
    """

    children: (
        list[
            paragraph.Paragraph
            | table.Table
            | Coroutine[None, None, paragraph.Paragraph | table.Table]
        ]
        | None
    ) = None


@dataclasses.dataclass
class Section(base.Component):
    """A document section with optional headers and footers.

    Attributes:
        children: List of Paragraph or Table components, or coroutines that
            resolve to these types.
        properties: Section configuration.
        headers: Dictionary mapping header types ('default', 'first', 'even')
            to Header components.
        footers: Dictionary mapping footer types ('default', 'first', 'even')
            to Footer components.
    """

    children: (
        list[
            paragraph.Paragraph
            | table.Table
            | Coroutine[None, None, paragraph.Paragraph | table.Table]
        ]
        | None
    ) = None
    properties: SectionProperties | None = None
    headers: dict[HeaderFooterType, Header] | None = None
    footers: dict[HeaderFooterType, Footer] | None = None



"""Declarative API for creating Word documents."""

# sergey: disable-file: IMP001 # Allow barrel exports.

from cmi_docx.declarative.base import Component
from cmi_docx.declarative.document import Document
from cmi_docx.declarative.image import ImageRun
from cmi_docx.declarative.paragraph import Break, Paragraph, Tab, TextRun
from cmi_docx.declarative.section import Footer, Header, Section, SectionProperties
from cmi_docx.declarative.table import Table, TableCell, TableRow

__all__ = [
    "Break",
    "Component",
    "Document",
    "Footer",
    "Header",
    "ImageRun",
    "Paragraph",
    "Section",
    "SectionProperties",
    "Tab",
    "Table",
    "TableCell",
    "TableRow",
    "TextRun",
]



"""Extends a python-docx Word document with additional functionality."""

import pathlib

from docx import document
from docx.text import paragraph as docx_paragraph

from cmi_docx import paragraph, run, styles


class ExtendDocument:
    """Extends a python-docx Word document with additional functionality."""

    def __init__(self, document: document.Document) -> None:
        """Initializes a DocxSearch object for finding text."""
        self.document = document

    def find_in_paragraphs(self, needle: str) -> list[paragraph.FindParagraph]:
        """Finds the indices of a text relative to the paragraphs.

        Args:
            needle: The text to find.

        Returns:
            The indices of the text in the document.
        """
        return [
            paragraph.ExtendParagraph(para).find_in_paragraph(needle)
            for para in self.all_paragraphs
        ]

    def find_in_runs(self, needle: str) -> list[run.FindRun]:
        """Finds the indices of a text relative to the document's runs.

        Args:
            needle: The text to find.

        Returns:
            The locations of the text in the runs.
        """
        return [
            finder
            for para in self.all_paragraphs
            for finder in paragraph.ExtendParagraph(para).find_in_runs(needle)
        ]

    def replace(
        self, needle: str, replace: str, style: styles.RunStyle | None = None
    ) -> None:
        """Finds and replaces text in a Word document.

        Args:
            needle: The text to find.
            replace: The text to replace.
            style: The style to apply to the replacement text.

        """
        run_finder = self.find_in_runs(needle)
        run_finder.sort(
            key=lambda x: (x.run_indices[0], x.character_indices[0]), reverse=True
        )

        for run_find in run_finder:
            run_find.replace(replace, style)

    def insert_paragraph_by_text(
        self,
        index: int,
        text: str,
        style: str | None = None,
    ) -> docx_paragraph.Paragraph:
        """Inserts a paragraph into a document.

        Args:
            index: The index to insert the paragraph at.
            text: The text to insert.
            style: The style to apply to the text.

        Returns:
            The new paragraph.
        """
        new_paragraph = self._insert_empty_paragraph(index, style)
        new_paragraph.add_run(text)
        return new_paragraph

    def insert_paragraph_by_object(
        self,
        index: int,
        paragraph: docx_paragraph.Paragraph,
    ) -> docx_paragraph.Paragraph:
        """Inserts a paragraph into a document.

        Args:
            index: The index to insert the paragraph at.
            paragraph: The paragraph to insert.

        Returns:
            The new paragraph.
        """
        new_paragraph = self._insert_empty_paragraph(index)
        for paragraph_run in paragraph.runs:
            new_paragraph.add_run(paragraph_run.text)
        return new_paragraph

    def insert_image(
        self,
        index: int,
        image_path: str | pathlib.Path,
        width: int | None = None,
        height: int | None = None,
    ) -> docx_paragraph.Paragraph:
        """Inserts an image at a given paragraph index.

        Args:
            index: The paragraph index to insert the image at.
            image_path: The path to the image to insert.
            width: The width of the image.
            height: The height of the image.
        """
        new_paragraph = self._insert_empty_paragraph(index)
        run = new_paragraph.add_run()
        run.add_picture(str(image_path), width=width, height=height)
        return new_paragraph

    @property
    def all_paragraphs(self) -> list[docx_paragraph.Paragraph]:
        """Returns all paragraphs in the document, including headers and footers."""
        all_paragraphs = list(self.document.paragraphs)

        for section in self.document.sections:
            all_paragraphs.extend(
                (*section.footer.paragraphs, *section.header.paragraphs)
            )
        return all_paragraphs

    def _insert_empty_paragraph(
        self, index: int, style: str | None = None
    ) -> docx_paragraph.Paragraph:
        """Inserts an empty paragraph at a given index.

        Args:
            index: The index to insert the paragraph at.
            style: The style to apply to the paragraph.

        Returns:
            The new paragraph.

        Raises:
            IndexError: If the index is out of range.
        """
        n_paragraphs = len(self.document.paragraphs)
        if index > n_paragraphs:
            msg = f"Index {index} is out of range."
            raise IndexError(msg)

        if index == n_paragraphs:
            new_paragraph = self.document.add_paragraph(style=style)
        else:
            new_paragraph = self.document.paragraphs[index]._insert_paragraph_before()  # noqa: SLF001
            new_paragraph.style = style

        return new_paragraph



"""Top-level Document class for declarative API."""

import asyncio
import dataclasses
import io
import pathlib

import docx
from docx import document as docx_document
from docx import section as docx_section
from docx import shared
from docx import table as docx_table
from docx.enum import section as docx_enum_section
from docx.enum import text as docx_text
from docx.text import paragraph as docx_paragraph

from cmi_docx import document as imperative_document
from cmi_docx.declarative import image, paragraph, section, table


@dataclasses.dataclass
class DocumentTemplate:
    """Defines a template to use as a base for a document.

    Attributes:
        path: Path to the document file.
        replacements: Dictionary of needle/replacements to make in the template.
    """

    path: pathlib.Path | str
    replacements: dict[str, str] | None = None


class Document:
    """A Word document with sections.

    This is the top-level container for a declarative document. All operations
    are async - use await to resolve all async children concurrently.

    Attributes:
        sections: List of Section components or coroutines that resolve to sections.
        creator: Document creator metadata.
        title: Document title metadata.
        subject: Document subject metadata.
        description: Document description metadata.
        keywords: Document keywords metadata.
        category: Document category metadata.
        comments: Document comments metadata.
        comment_author: Default author name for all Word comments. Can be overridden
            per-paragraph or per-text-run.
        styles: Document-level style definitions.
        numbering: Document-level numbering definitions.

    Example:
        >>> async def create_doc():
        ...     doc = Document(sections=[
        ...         Section(children=[
        ...             Paragraph(text="Hello World"),
        ...             fetch_paragraph(),  # async function
        ...         ]),
        ...     ])
        ...     await doc.save("output.docx")
    """

    def __init__(  # noqa: PLR0913, D107
        self,
        sections: list[section.Section],
        creator: str | None = None,
        title: str | None = None,
        subject: str | None = None,
        description: str | None = None,
        keywords: str | None = None,
        category: str | None = None,
        comments: str | None = None,
        comment_author: str | None = None,
        styles: dict[str, str | int | bool] | None = None,
        numbering: dict[str, str | int | list[dict[str, str | int]]] | None = None,
    ) -> None:
        self.sections = sections
        self.creator = creator
        self.title = title
        self.subject = subject
        self.description = description
        self.keywords = keywords
        self.category = category
        self.comments = comments
        self.comment_author = comment_author
        self.styles = styles
        self.numbering = numbering

    async def to_docx(  # noqa: C901
        self, template: DocumentTemplate | None = None
    ) -> docx_document.Document:
        """Convert to a python-docx Document.

        Automatically resolves all async children before converting.

        Returns:
            A python-docx Document object.
        """
        await asyncio.gather(*(sec.resolve() for sec in self.sections))
        if template is not None:
            docx_doc = docx.Document()
            if template.replacements is not None:
                extended_doc = imperative_document.ExtendDocument(docx_doc)
                for needle, replacement in template.replacements.items():
                    extended_doc.replace(needle, replacement)
        else:
            docx_doc = (
                docx.Document()
                if template is None
                else docx.Document(str(template.path))
            )

        if self.creator:
            docx_doc.core_properties.author = self.creator
        if self.title:
            docx_doc.core_properties.title = self.title
        if self.subject:
            docx_doc.core_properties.subject = self.subject
        if self.description:
            docx_doc.core_properties.comments = self.description
        if self.keywords:
            docx_doc.core_properties.keywords = self.keywords
        if self.category:
            docx_doc.core_properties.category = self.category

        for sec in self.sections:
            _pack_section(docx_doc, sec, self.comment_author)

        return docx_doc


def _pack_section(  # noqa: C901, PLR0912
    docx_doc: docx_document.Document,
    sec: section.Section,
    default_comment_author: str | None,
) -> None:
    """Pack a Section into a python-docx document.

    Args:
        docx_doc: The python-docx Document.
        sec: The declarative Section.
        default_comment_author: Default author for comments.
    """
    if not sec.condition():
        return

    if sec.children:
        for child in sec.children:
            _pack_block_element(docx_doc, docx_doc, child, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.

    docx_section = docx_doc.add_section()

    if sec.properties:
        props = sec.properties
        if props.page_size:
            if "width" in props.page_size:
                docx_section.page_width = props.page_size["width"]
            if "height" in props.page_size:
                docx_section.page_height = props.page_size["height"]

        if props.page_orientation:
            if props.page_orientation.lower() == "landscape":
                docx_section.orientation = docx_enum_section.WD_ORIENTATION.LANDSCAPE
            elif props.page_orientation.lower() == "portrait":
                docx_section.orientation = docx_enum_section.WD_ORIENTATION.PORTRAIT

    if sec.headers:
        for header_type, header in sec.headers.items():
            _pack_header(
                docx_doc, docx_section, header_type, header, default_comment_author
            )

    if sec.footers:
        for footer_type, footer in sec.footers.items():
            _pack_footer(
                docx_doc, docx_section, footer_type, footer, default_comment_author
            )


def _get_header_or_footer(
    section: docx_section.Section, hf_type: str, *, is_header: bool
) -> docx_section.Section | None:
    """Get the appropriate header or footer from a section.

    Args:
        section: The python-docx Section.
        hf_type: The type ('default', 'first', 'even').
        is_header: True for header, False for footer.

    Returns:
        The header or footer object, or None if type is invalid.
    """
    type_mapping = {
        "default": "header" if is_header else "footer",
        "first": "first_page_header" if is_header else "first_page_footer",
        "even": "even_page_header" if is_header else "even_page_footer",
    }

    attr_name = type_mapping.get(hf_type)
    return getattr(section, attr_name, None) if attr_name else None


def _pack_header(
    docx_doc: docx_document.Document,
    section: docx_section.Section,
    header_type: str,
    header: section.Header,
    default_comment_author: str | None,
) -> None:
    """Pack a Header into a python-docx section.

    Args:
        docx_doc: The python-docx Document.
        section: The python-docx Section.
        header_type: The header type ('default', 'first', 'even').
        header: The declarative Header.
        default_comment_author: Default author for comments.
    """
    if not header.condition():
        return

    docx_header = _get_header_or_footer(section, header_type, is_header=True)
    if docx_header and header.children:
        for child in header.children:
            _pack_block_element(docx_doc, docx_header, child, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.


def _pack_footer(
    docx_doc: docx_document.Document,
    section: docx_section.Section,
    footer_type: str,
    footer: section.Footer,
    default_comment_author: str | None,
) -> None:
    """Pack a Footer into a python-docx section.

    Args:
        docx_doc: The python-docx Document.
        section: The python-docx Section.
        footer_type: The footer type ('default', 'first', 'even').
        footer: The declarative Footer.
        default_comment_author: Default author for comments.
    """
    if not footer.condition():
        return

    docx_footer = _get_header_or_footer(section, footer_type, is_header=False)
    if docx_footer and footer.children:
        for child in footer.children:
            _pack_block_element(docx_doc, docx_footer, child, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.


def _pack_block_element(
    docx_doc: docx_document.Document,
    container: docx_document.Document,
    element: paragraph.Paragraph | table.Table,
    default_comment_author: str | None,
) -> None:
    """Pack a block-level element (Paragraph or Table).

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        container: The container to add to.
        element: The Paragraph or Table to pack.
        default_comment_author: Default author for comments.
    """
    if not element.condition():
        return None

    if isinstance(element, paragraph.Paragraph):
        return _pack_paragraph(docx_doc, container, element, default_comment_author)
    return _pack_table(docx_doc, container, element, default_comment_author)


def _pack_paragraph(
    docx_doc: docx_document.Document,
    container: docx_document.Document,
    para: paragraph.Paragraph,
    default_comment_author: str | None,
) -> None:
    """Pack a Paragraph into a container.

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        container: The container to add to.
        para: The declarative Paragraph.
        default_comment_author: Default author for comments.
    """
    docx_para = container.add_paragraph()
    _pack_paragraph_into_existing(docx_doc, docx_para, para, default_comment_author)


def _pack_paragraph_into_existing(  # noqa: C901, PLR0912
    docx_doc: docx_document.Document,
    docx_para: docx_paragraph.Paragraph,
    para: paragraph.Paragraph,
    default_comment_author: str | None,
) -> None:
    """Pack a Paragraph into an existing python-docx paragraph.

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        docx_para: The python-docx Paragraph to populate.
        para: The declarative Paragraph.
        default_comment_author: Default author for comments.
    """
    if para.style:
        docx_para.style = para.style

    if para.heading:
        docx_para.style = f"Heading {para.heading}"

    if para.alignment:
        docx_para.alignment = para.alignment

    fmt = docx_para.paragraph_format
    if para.spacing_before:
        fmt.space_before = shared.Pt(para.spacing_before)
    if para.spacing_after:
        fmt.space_after = shared.Pt(para.spacing_after)
    if para.line_spacing:
        fmt.line_spacing = para.line_spacing
    if para.left_indent:
        fmt.left_indent = shared.Pt(para.left_indent)
    if para.right_indent:
        fmt.right_indent = shared.Pt(para.right_indent)
    if para.first_line_indent:
        fmt.first_line_indent = shared.Pt(para.first_line_indent)
    if para.keep_together is not None:
        fmt.keep_together = para.keep_together
    if para.keep_with_next is not None:
        fmt.keep_with_next = para.keep_with_next
    if para.page_break_before is not None:
        fmt.page_break_before = para.page_break_before
    if para.widow_control is not None:
        fmt.widow_control = para.widow_control

    if para.text:
        docx_para.add_run(para.text)  # ty:ignore[invalid-argument-type] Text is already awaited.
    elif para.children:
        for child in para.children:
            _pack_inline_element(docx_doc, docx_para, child, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.

    if para.comment_text:
        author = para.comment_author or default_comment_author or ""
        docx_doc.add_comment(runs=docx_para.runs, text=para.comment_text, author=author)  # ty:ignore[invalid-argument-type] already awaited.


def _pack_inline_element(
    docx_doc: docx_document.Document,
    para: docx_paragraph.Paragraph,
    element: paragraph.TextRun | image.ImageRun | paragraph.Tab | paragraph.Break,
    default_comment_author: str | None,
) -> None:
    """Pack an inline element (TextRun, ImageRun, Tab, Break).

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        para: The python-docx Paragraph.
        element: The inline element to pack.
        default_comment_author: Default author for comments.
    """
    if not element.condition():
        return

    if isinstance(element, paragraph.TextRun):
        _pack_text_run(docx_doc, para, element, default_comment_author)
    elif isinstance(element, image.ImageRun):
        _pack_image_run(para, element)
    elif isinstance(element, paragraph.Tab):
        para.add_run().add_tab()
    elif isinstance(element, paragraph.Break):
        break_type_mapping = {
            "page": docx_text.WD_BREAK.PAGE,
            "column": docx_text.WD_BREAK.COLUMN,
        }
        break_type = break_type_mapping.get(element.type)
        if break_type:
            para.add_run().add_break(break_type)
        else:
            para.add_run().add_break()


def _pack_text_run(  # noqa: C901
    docx_doc: docx_document.Document,
    para: docx_paragraph.Paragraph,
    run: paragraph.TextRun,
    default_comment_author: str | None,
) -> None:
    """Pack a TextRun into a paragraph.

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        para: The python-docx Paragraph.
        run: The declarative TextRun.
        default_comment_author: Default author for comments.
    """
    docx_run = para.add_run(run.text)  # ty:ignore[invalid-argument-type] Already awaited.

    if run.bold is not None:
        docx_run.bold = run.bold
    if run.italic is not None:
        docx_run.italic = run.italic
    if run.underline is not None:
        docx_run.underline = run.underline

    font = docx_run.font
    if run.font:
        font.name = run.font
    if run.size:
        font.size = shared.Pt(run.size)
    if run.color:
        font.color.rgb = shared.RGBColor(*run.color)
    if run.superscript:
        font.superscript = True
    if run.subscript:
        font.subscript = True
    if run.strike:
        font.strike = True
    if run.all_caps:
        font.all_caps = True
    if run.small_caps:
        font.small_caps = True

    if run.comment_text:
        author = run.comment_author or default_comment_author or ""
        docx_doc.add_comment(runs=docx_run, text=run.comment_text, author=author)  # ty:ignore[invalid-argument-type] already awaited.


def _pack_image_run(para: docx_paragraph.Paragraph, img: image.ImageRun) -> None:
    """Pack an ImageRun into a paragraph.

    Args:
        para: The python-docx Paragraph.
        img: The declarative ImageRun.
    """
    docx_run = para.add_run()

    width = None
    height = None
    if img.transformation:
        if "width" in img.transformation:
            width = shared.Pt(img.transformation["width"])
        if "height" in img.transformation:
            height = shared.Pt(img.transformation["height"])

    if isinstance(img.data, bytes):
        docx_run.add_picture(
            io.BytesIO(img.data),
            width=width,
            height=height,
        )
    elif isinstance(img.data, (str, pathlib.Path)):
        docx_run.add_picture(
            str(img.data),
            width=width,
            height=height,
        )


def _pack_table(
    docx_doc: docx_document.Document,
    container: docx_document.Document,
    tbl: table.Table,
    default_comment_author: str | None,
) -> None:
    """Pack a Table into a container.

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        container: The container to add to.
        tbl: The declarative Table.
        default_comment_author: Default author for comments.
    """
    filtered_rows = [row for row in tbl.rows if row.condition()]  # ty:ignore[unresolved-attribute] already awaited.
    if not filtered_rows:
        return

    num_rows = len(filtered_rows)
    num_cols = (
        max(
            len([cell for cell in row.children if cell.condition()])  # ty:ignore[unresolved-attribute] already awaited.
            for row in filtered_rows
        )
        if filtered_rows
        else 0
    )

    docx_table = container.add_table(rows=num_rows, cols=num_cols)

    if tbl.style:
        docx_table.style = tbl.style

    for row_idx, row in enumerate(filtered_rows):
        _pack_table_row(docx_doc, docx_table.rows[row_idx], row, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.


def _pack_table_row(
    docx_doc: docx_document.Document,
    docx_row: docx_table._Row,
    row: table.TableRow,
    default_comment_author: str | None,
) -> None:
    """Pack a TableRow.

    Args:
        docx_doc: The python-docx Document.
        docx_row: The python-docx table row.
        row: The declarative TableRow.
        default_comment_author: Default author for comments.
    """
    filtered_cells = [cell for cell in row.children if cell.condition()]  # ty:ignore[unresolved-attribute] already awaited.
    for cell_idx, cell in enumerate(filtered_cells):
        _pack_table_cell(
            docx_doc,
            docx_row.cells[cell_idx],
            cell,  # ty:ignore[invalid-argument-type] Cell already awaited.
            default_comment_author,
        )


def _pack_table_cell(
    docx_doc: docx_document.Document,
    docx_cell: docx_table._Cell,
    cell: table.TableCell,
    default_comment_author: str | None,
) -> None:
    """Pack a TableCell.

    Args:
        docx_doc: The python-docx Document (needed for comment API).
        docx_cell: The python-docx table cell.
        cell: The declarative TableCell.
        default_comment_author: Default author for comments.
    """
    if cell.children:
        for idx, child in enumerate(cell.children):
            if idx == 0 and isinstance(child, paragraph.Paragraph):
                _pack_paragraph_into_existing(
                    docx_doc, docx_cell.paragraphs[0], child, default_comment_author
                )
            else:
                _pack_block_element(docx_doc, docx_cell, child, default_comment_author)  # ty:ignore[invalid-argument-type] already awaited.



"""Style interfaces for document properties."""

import dataclasses
from typing import Literal

from docx.enum import text


@dataclasses.dataclass
class RunStyle:
    """Dataclass for run style arguments."""

    bold: bool | None = None
    italic: bool | None = None
    underline: bool | None = None
    superscript: bool | None = None
    subscript: bool | None = None
    font_size: int | None = None
    font_rgb: tuple[int, int, int] | None = None


@dataclasses.dataclass
class ParagraphStyle:
    """Dataclass for paragraph style arguments."""

    bold: bool | None = None
    italic: bool | None = None
    font_size: int | None = None
    font_rgb: tuple[int, int, int] | None = None
    line_spacing: float | None = None
    space_before: int | None = None
    space_after: int | None = None
    alignment: text.WD_PARAGRAPH_ALIGNMENT | None = None


@dataclasses.dataclass
class CellBorder:
    """Dataclass for cell border style arguments.

    Attributes:
        sides: Tuple of sides for this to apply to, must be "top", "bottom", "insideH",
            "insideV", "start", or "end".
        sz: Size of the border, 1pt = 8/
        val: Whether to use a single or dashed line. None makes the line invisible.
        color: The color of the border as a hex code.
    """

    sides: tuple[str, ...]
    sz: int | None = None
    val: Literal["single", "dashed"] | None = "single"
    color: str | None = None


@dataclasses.dataclass
class CellStyle:
    """Dataclass for table style arguments."""

    paragraph: ParagraphStyle | None = None
    background_rgb: tuple[int, int, int] | None = None
    borders: list[CellBorder] | None = None


@dataclasses.dataclass
class TableSections:
    """Dataclass for enabling/disabling sections of a Word table."""

    first_column: bool | None = None
    first_row: bool | None = None
    last_column: bool | None = None
    last_row: bool | None = None
    no_h_band: bool | None = None
    no_v_band: bool | None = None

    # Alias snake case Python style to camelCase used in Word format.
    @property
    def firstColumn(self) -> bool | None:  # noqa: N802
        """Alias for first_column."""
        return self.first_column

    @property
    def firstRow(self) -> bool | None:  # noqa: N802
        """Alias for first_row."""
        return self.first_row

    @property
    def lastColumn(self) -> bool | None:  # noqa: N802
        """Alias for last_column."""
        return self.last_column

    @property
    def lastRow(self) -> bool | None:  # noqa: N802
        """Alias for last_row."""
        return self.last_row

    @property
    def noHBand(self) -> bool | None:  # noqa: N802
        """Alias for no_h_band."""
        return self.no_h_band

    @property
    def noVBand(self) -> bool | None:  # noqa: N802
        """Alias for no_v_band."""
        return self.no_v_band


@dataclasses.dataclass
class TableStyle:
    """Dataclass for table style arguments."""

    sections: TableSections | None = None



"""Extends a python-docx Table cell with additional functionality."""

from docx import oxml, table
from docx.oxml import ns

from cmi_docx import paragraph, styles


class ExtendTable:
    """Extends a python-docx Table with additional functionality."""

    def __init__(self, tbl: table.Table) -> None:
        """Initialize the ExtendTable object.

        Args:
            tbl: The table to extend.
        """
        self.table = tbl

    def format(self, style: styles.TableStyle) -> None:
        """Formats a table in a Word document.

        Args:
            style: The style to use.

        Raises:
            ValueError: If table look was not found.
        """
        if style.sections:
            section_names = (
                "firstColumn",
                "firstRow",
                "lastColumn",
                "lastRow",
                "noHBand",
                "noVBand",
            )
            for name in section_names:
                value = getattr(style.sections, name)
                if value is None:
                    continue

                tbl_pr = self.table._tblPr  # noqa: SLF001
                tbl_look = tbl_pr.first_child_found_in("w:tblLook")
                if tbl_look is None:
                    msg = "Table look was not found."
                    raise ValueError(msg)
                tbl_look.set(ns.qn(f"w:{name}"), str(int(value)))


class ExtendCell:
    """Extends a python-docx Word cell with additional functionality."""

    def __init__(self, cell: table._Cell) -> None:
        """Initializes an ExtendCell object.

        Args:
            cell: The cell to extend.
        """
        self.cell = cell

    def format(self, style: styles.CellStyle) -> None:
        """Formats a cell in a Word table.

        Args:
            style: The style to apply to the cell.
        """
        if style.paragraph is not None:
            for table_paragraph in self.cell.paragraphs:
                paragraph.ExtendParagraph(table_paragraph).format(style.paragraph)

        if style.background_rgb is not None:
            shading = oxml.parse_xml(
                (
                    r'<w:shd {} w:fill="'
                    f"{rgb_to_hex(*style.background_rgb)}"
                    r'"/>'
                ).format(
                    ns.nsdecls("w"),
                ),
            )
            self.cell._tc.get_or_add_tcPr().append(shading)  # noqa: SLF001

        if style.borders:
            for border in style.borders:
                self._apply_border(border)

    def _apply_border(self, border: styles.CellBorder) -> None:
        """Applies the borders styling to the cell.

        Args:
            border: The style to apply to the cell.
        """
        tc_pr = self.cell._tc.get_or_add_tcPr()  # noqa: SLF001

        tc_borders = tc_pr.first_child_found_in("w:tcBorders")
        if tc_borders is None:
            tc_borders = oxml.OxmlElement("w:tcBorders")
            tc_pr.append(tc_borders)

        for edge in border.sides:
            tag = f"w:{edge}"
            element = tc_borders.find(ns.qn(tag))
            if element is None:
                element = oxml.OxmlElement(tag)
                tc_borders.append(element)

            # looks like order of attributes is important
            for key in ["sz", "val", "color"]:
                if value := getattr(border, key):
                    element.set(ns.qn(f"w:{key}"), str(value))


def rgb_to_hex(red: int, green: int, blue: int) -> str:
    """Converts RGB values to a hexadecimal color code.

    Args:
        red: The red component of the RGB color.
        green: The green component of the RGB color.
        blue: The blue component of the RGB color.

    Returns:
        The hexadecimal color code representing the RGB color.
    """
    return f"#{red:02x}{green:02x}{blue:02x}".upper()



"""Tests for conditional rendering in the declarative API."""

import pytest

from cmi_docx import declarative


@pytest.mark.asyncio
async def test_section_condition_false() -> None:
    """Test that a section with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Visible section"),
                ],
            ),
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Hidden section"),
                ],
                condition=lambda: False,
            ),
        ],
    )

    docx = await doc.to_docx()
    assert len(docx.sections) == 2  # noqa: PLR2004
    content_paragraphs = [p for p in docx.paragraphs if p.text.strip()]
    assert len(content_paragraphs) == 1
    assert content_paragraphs[0].text.startswith("Visible section")


@pytest.mark.asyncio
async def test_paragraph_condition_false() -> None:
    """Test that a paragraph with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="First paragraph"),
                    declarative.Paragraph(
                        text="Hidden paragraph", condition=lambda: False
                    ),
                    declarative.Paragraph(text="Third paragraph"),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    content_paragraphs = [p for p in docx.paragraphs if p.text.strip()]
    assert len(content_paragraphs) == 2  # noqa: PLR2004
    assert content_paragraphs[0].text.startswith("First paragraph")
    assert content_paragraphs[1].text.startswith("Third paragraph")


@pytest.mark.asyncio
async def test_text_run_condition_false() -> None:
    """Test that a text run with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(
                        children=[
                            declarative.TextRun(text="First "),
                            declarative.TextRun(
                                text="Hidden ", condition=lambda: False
                            ),
                            declarative.TextRun(text="Third"),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert len(docx.paragraphs[0].runs) == 2  # noqa: PLR2004
    assert docx.paragraphs[0].text == "First Third"


@pytest.mark.asyncio
async def test_table_row_condition_false() -> None:
    """Test that a table row with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Table(
                        rows=[
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Row 1"),
                                        ]
                                    ),
                                ],
                            ),
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(
                                                text="Row 2 - Hidden"
                                            ),
                                        ]
                                    ),
                                ],
                                condition=lambda: False,
                            ),
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Row 3"),
                                        ]
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert len(docx.tables[0].rows) == 2  # noqa: PLR2004
    assert docx.tables[0].rows[0].cells[0].paragraphs[0].text == "Row 1"
    assert docx.tables[0].rows[1].cells[0].paragraphs[0].text == "Row 3"


@pytest.mark.asyncio
async def test_table_cell_condition_false() -> None:
    """Test that a table cell with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Table(
                        rows=[
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Cell 1"),
                                        ]
                                    ),
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(
                                                text="Cell 2 - Hidden"
                                            ),
                                        ],
                                        condition=lambda: False,
                                    ),
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Cell 3"),
                                        ]
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert len(docx.tables[0].rows[0].cells) == 2  # noqa: PLR2004
    assert docx.tables[0].rows[0].cells[0].paragraphs[0].text == "Cell 1"
    assert docx.tables[0].rows[0].cells[1].paragraphs[0].text == "Cell 3"


@pytest.mark.asyncio
async def test_table_condition_false() -> None:
    """Test that a table with condition=False is not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Before table"),
                    declarative.Table(
                        rows=[
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Hidden table"),
                                        ]
                                    ),
                                ],
                            ),
                        ],
                        condition=lambda: False,
                    ),
                    declarative.Paragraph(text="After table"),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert len(docx.tables) == 0
    content_paragraphs = [p for p in docx.paragraphs if p.text.strip()]
    assert len(content_paragraphs) == 2  # noqa: PLR2004
    assert content_paragraphs[0].text.startswith("Before table")
    assert content_paragraphs[1].text.startswith("After table")


@pytest.mark.asyncio
async def test_nested_condition_false() -> None:
    """Test that children of false-conditioned components are not rendered."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Visible"),
                    declarative.Paragraph(
                        children=[
                            declarative.TextRun(text="This whole paragraph is hidden"),
                        ],
                        condition=lambda: False,
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    content_paragraphs = [p for p in docx.paragraphs if p.text.strip()]
    assert len(content_paragraphs) == 1
    assert content_paragraphs[0].text.startswith("Visible")



"""Table tests for the declarative API."""

import pytest

from cmi_docx import declarative


@pytest.mark.asyncio
async def test_simple_table() -> None:
    """Test creating a document with a simple table."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Table(
                        rows=[
                            declarative.TableRow(
                                children=[
                                    declarative.TableCell(
                                        children=[
                                            declarative.Paragraph(text="Header 1"),
                                            declarative.Paragraph(text="Paragraph 2"),
                                        ]
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.tables[0].rows[0].cells[0].paragraphs[0].text == "Header 1"
    assert docx.tables[0].rows[0].cells[0].paragraphs[1].text == "Paragraph 2"



"""Paragraph and text run components for declarative documents."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from cmi_docx.declarative import base

if TYPE_CHECKING:
    from collections.abc import Awaitable, Coroutine

    from docx.enum import text as docx_text

    from cmi_docx.declarative import image

InlineElement = "TextRun | image.ImageRun | Tab | Break"


@dataclasses.dataclass
class TextRun(base.Component):
    """A run of text with formatting.

    Attributes:
        text: The text content.
        comment_text: Text content for a Word comment attached to this run.
        comment_author: Author name for the comment. Overrides the
            document-level default.
        bold: Apply bold formatting.
        italic: Apply italic formatting.
        underline: Apply underline formatting.
        font: Font name.
        size: Font size in points.
        color: Text color as (R, G, B) tuple (0-255).
        superscript: Apply superscript formatting.
        subscript: Apply subscript formatting.
        strike: Apply strikethrough formatting.
        all_caps: Apply all caps formatting.
        small_caps: Apply small caps formatting.
    """

    text: Awaitable[str] | str
    comment_text: Awaitable[str] | str | None = None
    comment_author: Awaitable[str] | str | None = None
    bold: bool | None = None
    italic: bool | None = None
    underline: bool | None = None
    font: str | None = None
    size: int | None = None
    color: tuple[int, int, int] | None = None
    superscript: bool | None = None
    subscript: bool | None = None
    strike: bool | None = None
    all_caps: bool | None = None
    small_caps: bool | None = None


@dataclasses.dataclass
class Tab(base.Component):
    """A tab character."""


@dataclasses.dataclass
class Break(base.Component):
    """A line or page break.

    Attributes:
        type: Type of break ('line', 'page', 'column', 'textWrapping').
    """

    type: str = "line"


@dataclasses.dataclass
class Paragraph(base.Component):
    """A paragraph with optional formatting and child runs.

    Attributes:
        text: Shorthand for a paragraph with a single text run.
        comment_text: Text content for a Word comment attached to this paragraph.
        comment_author: Author name for the comment. Overrides the
            document-level default.
        children: List of TextRun, ImageRun, Tab, Break, or coroutines
            that resolve to these types.
        heading: Heading level.
        style: Style name to apply.
        alignment: Paragraph alignment.
        spacing_before: Space before paragraph in points.
        spacing_after: Space after paragraph in points.
        line_spacing: Line spacing multiplier.
        left_indent: Left indent in points.
        right_indent: Right indent in points.
        first_line_indent: First line indent in points.
        keep_together: Keep all lines of paragraph on same page.
        keep_with_next: Keep paragraph with next paragraph.
        page_break_before: Start paragraph on new page.
        widow_control: Enable widow/orphan control.
    """

    text: Awaitable[str] | str | None = None
    comment_text: Awaitable[str] | str | None = None
    comment_author: Awaitable[str] | str | None = None
    children: (
        list[
            TextRun
            | image.ImageRun
            | Tab
            | Break
            | Coroutine[None, None, TextRun | image.ImageRun | Tab | Break]
        ]
        | None
    ) = None
    heading: int | None = None
    style: str | None = None
    alignment: docx_text.WD_PARAGRAPH_ALIGNMENT | None = None
    spacing_before: int | None = None
    spacing_after: int | None = None
    line_spacing: float | None = None
    left_indent: int | None = None
    right_indent: int | None = None
    first_line_indent: int | None = None
    keep_together: bool | None = None
    keep_with_next: bool | None = None
    page_break_before: bool | None = None
    widow_control: bool | None = None

    def __post_init__(self) -> None:
        """Validate that either text or children is provided, not both.

        Raises:
            ValueError: If both text and children are provided, or neither is provided.
        """
        if self.text is not None and self.children is not None:
            msg = "Paragraph cannot have both 'text' and 'children'"
            raise ValueError(msg)
        if self.text is None and self.children is None:
            msg = "Paragraph must have either 'text' or 'children'"
            raise ValueError(msg)



"""Tests for the comment module."""

import docx
from docx.opc import constants as docx_constants

import cmi_docx
from cmi_docx import comment


def test_add_comment_single() -> None:
    """Tests adding a comment to a single entity."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    author = "Grievous"
    message = "Ah, General Kenobi."

    cmi_docx.add_comment(document, para, author, message)
    para_comment = para.part.part_related_by(docx_constants.RELATIONSHIP_TYPE.COMMENTS)

    assert author in para_comment.blob.decode()
    assert message in para_comment.blob.decode()


def test_add_comment_range() -> None:
    """Tests adding a comment with a range of entities."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    run_start = para.add_run(" run 3!")
    run_end = para.add_run(" run 4!")
    author = "Grievous"
    message = "Ah, General Kenobi."

    cmi_docx.add_comment(document, (run_start, run_end), author, message)
    comment_start = run_start.part.part_related_by(
        docx_constants.RELATIONSHIP_TYPE.COMMENTS
    )
    comment_end = run_end.part.part_related_by(
        docx_constants.RELATIONSHIP_TYPE.COMMENTS
    )

    assert author in comment_start.blob.decode()
    assert message in comment_start.blob.decode()
    assert author in comment_end.blob.decode()
    assert message in comment_end.blob.decode()


def test_comment_preserver_extract_comments() -> None:
    """Tests getting comments from a paragraph."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    author = "Grievous"
    message = "Ah, General Kenobi."
    cmi_docx.add_comment(document, para, author, message)
    preserver = comment.CommentPreserver(para._element)

    comments = preserver.extract_comments()

    assert len(comments) == 1
    assert comments[0].start_index == 0
    assert comments[0].end_index == len(para.text)


def test_comment_preserver_strip_comments() -> None:
    """Tests stripping comments from a paragraph."""
    document = docx.Document()
    para = document.add_paragraph()
    run = para.add_run("this")
    para.add_run("is")
    cmi_docx.add_comment(document, run, "Grievous", "Ah, General Kenobi.")
    preserver = comment.CommentPreserver(para._element)

    preserver.strip_comments()

    assert para.text == "thisis"


def test_replace_between_preserve_comments_contained() -> None:
    """Using the replace_between with an edit inside the comment."""
    document = docx.Document()
    para = document.add_paragraph("This is a sample paragraph.")
    extend_para = cmi_docx.ExtendParagraph(para)
    author = "Grievous"
    message = "Ah, General Kenobi."
    cmi_docx.add_comment(document, para, author, message)

    extend_para.replace_between(5, 7, "was")
    comments = comment.CommentPreserver(para._element).extract_comments()

    assert para.text == "This was a sample paragraph."
    assert len(comments) == 1
    assert comments[0].start_index == 0
    assert comments[0].end_index == len(para.text)


def test_strip_comments_removes_comments_entire_string() -> None:
    """Using the replace_between with the edit equivalent to the comment range."""
    document = docx.Document()
    para = document.add_paragraph("")
    run = para.add_run("this")
    para.add_run("is")
    extend_para = cmi_docx.ExtendParagraph(para)
    cmi_docx.add_comment(document, run, "Grievous", "Ah, General Kenobi.")

    extend_para.replace_between(0, len(run.text), "was")
    comments = comment.CommentPreserver(para._element).extract_comments()

    assert para.text == "wasis"
    assert len(comments) == 1
    assert comments[0].start_index == 0
    assert comments[0].end_index == len(para.runs[0].text)


def test_restore_comments_preserves_comment_references() -> None:
    """Tests that restoring comments doesn't remove comment reference runs."""
    document = docx.Document()
    para = document.add_paragraph("Sample text with comment.")
    author = "Obi-Wan"
    message = "Hello there!"

    cmi_docx.add_comment(document, para, author, message)
    cmi_docx.add_comment(document, para, author, message)

    original_xml = para._element
    comment_refs_before = len(original_xml.xpath(".//w:commentReference"))

    preserver = comment.CommentPreserver(para._element)
    extracted_comments = preserver.extract_comments()
    preserver.restore_comments(extracted_comments)
    comment_refs_after = len(para._element.xpath(".//w:commentReference"))

    assert comment_refs_before > 0
    assert comment_refs_after == comment_refs_before
    assert para.text == "Sample text with comment."



"""Async tests for the declarative API."""

import asyncio

import pytest

from cmi_docx import declarative


async def fetch_paragraph() -> declarative.Paragraph:
    """Simulate fetching a paragraph asynchronously."""
    await asyncio.sleep(0.01)
    return declarative.Paragraph(text="Async paragraph")


async def fetch_text_run() -> declarative.TextRun:
    """Simulate fetching a text run asynchronously."""
    await asyncio.sleep(0.01)
    return declarative.TextRun(text="async text", bold=True)


@pytest.mark.asyncio
async def test_async_paragraph() -> None:
    """Test creating a document with async paragraphs."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(text="Sync paragraph"),
                    fetch_paragraph(),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.paragraphs[0].text.startswith("Sync")
    assert docx.paragraphs[1].text.startswith("Async")


@pytest.mark.asyncio
async def test_async_text_run() -> None:
    """Test creating a document with async text runs."""
    doc = declarative.Document(
        sections=[
            declarative.Section(
                children=[
                    declarative.Paragraph(
                        children=[
                            declarative.TextRun(text="Sync text"),
                            fetch_text_run(),
                        ],
                    ),
                ],
            ),
        ],
    )

    docx = await doc.to_docx()
    assert docx.paragraphs[0].text.startswith("Sync text")
    assert "async text" in docx.paragraphs[0].text



"""Tests for the run module."""

import docx
import pytest

from cmi_docx import run, styles


def test_find_run_lt_same_paragraph() -> None:
    """Test that comparing FindRun of the same paragraph works."""
    document = docx.Document()
    paragraph = document.add_paragraph("Hello, world!")

    find_run1 = run.FindRun(paragraph, (0, 1), (0, 5))
    find_run2 = run.FindRun(paragraph, (1, 2), (5, 10))

    assert find_run1 < find_run2


def test_find_run_lt_different_paragraphs() -> None:
    """Test that comparing FindRun of different paragraphs fails.

    Raises:
        ValueError: When comparing FindRuns from different paragraphs.
    """
    document = docx.Document()
    paragraph1 = document.add_paragraph("Hello, world!")
    paragraph2 = document.add_paragraph("Hello, world!")

    find_run1 = run.FindRun(paragraph1, (0, 1), (0, 5))
    find_run2 = run.FindRun(paragraph2, (0, 1), (0, 5))

    with pytest.raises(
        ValueError, match="Cannot compare FindRun objects from different paragraphs"
    ):
        assert find_run1 < find_run2


def test_find_run_replace_no_style_one_run() -> None:
    """Test that replacing a run without style works."""
    document = docx.Document()
    paragraph = document.add_paragraph("")
    paragraph.add_run("Hello, world!")

    find_run = run.FindRun(paragraph, (0, 1), (0, 5))
    find_run.replace("Goodbye")

    assert paragraph.text == "Goodbye, world!"


def test_find_run_replace_twice() -> None:
    """Test that replacing a run twice raises ValueError.

    Raises:
        ValueError: When attempting a second replacement.
    """
    document = docx.Document()
    paragraph = document.add_paragraph("")
    paragraph.add_run("Hello, world!")

    find_run = run.FindRun(paragraph, (0, 1), (0, 5))
    find_run.replace("Goodbye")

    with pytest.raises(
        ValueError, match="Cannot use a FindRun replacement more than once"
    ):
        find_run.replace("Goodbye")


def test_find_run_replace_no_style_multiple_runs() -> None:
    """Test that replacing multiple runs without style works."""
    document = docx.Document()
    paragraph = document.add_paragraph("")
    paragraph.add_run("Hello, ")
    paragraph.add_run("world!")

    find_run = run.FindRun(paragraph, (0, 2), (5, 6))
    find_run.replace("Goodbye")

    assert paragraph.text == "HelloGoodbye"


def test_find_run_replace_with_style() -> None:
    """Test that replacing a run with style works."""
    document = docx.Document()
    paragraph = document.add_paragraph("")
    paragraph.add_run("Hello, world!")

    find_run = run.FindRun(paragraph, (0, 1), (0, 5))
    find_run.replace("Goodbye", styles.RunStyle(bold=True))

    assert paragraph.text == "Goodbye, world!"
    assert not paragraph.runs[0].bold
    assert paragraph.runs[1].bold
    assert paragraph.runs[1].text == "Goodbye"


def test_extend_run_format() -> None:
    """Test that formatting a run works."""
    document = docx.Document()
    paragraph = document.add_paragraph("Hello, world!")
    paragraph_run = paragraph.runs[0]

    extend_run = run.ExtendRun(paragraph_run)
    extend_run.format(
        styles.RunStyle(
            bold=True,
            italic=True,
            underline=True,
            superscript=True,
            font_rgb=(1, 0, 0),
        )
    )

    assert paragraph_run.bold
    assert paragraph_run.italic
    assert paragraph_run.underline
    assert paragraph_run.font.superscript
    assert paragraph_run.font.color.rgb == (1, 0, 0)



"""Base class for declarative components with async resolution support."""

import asyncio
import dataclasses
from collections.abc import Awaitable, Callable, Generator
from typing import Any, Self


@dataclasses.dataclass
class _ResolveTask:
    field_name: str
    idx: int | None
    awaitable: Awaitable[Any]


def _collect_resolve_tasks(component: "Component") -> list[_ResolveTask]:
    """Collect all async resolution tasks from a component's fields.

    Args:
        component: The component to collect tasks from.

    Returns:
        List of resolution tasks for all async children.
    """
    tasks: list[_ResolveTask] = []
    for field in dataclasses.fields(component):
        value = getattr(component, field.name)
        if value is None:
            continue

        if isinstance(value, Component):
            tasks.append(_ResolveTask(field.name, None, value.resolve()))
        elif asyncio.iscoroutine(value) or asyncio.isfuture(value):
            tasks.append(_ResolveTask(field.name, None, value))
        elif isinstance(value, (list, tuple, Generator)):
            for idx, item in enumerate(value):
                if isinstance(item, Component):
                    tasks.append(_ResolveTask(field.name, idx, item.resolve()))
                elif asyncio.iscoroutine(item) or asyncio.isfuture(item):
                    tasks.append(_ResolveTask(field.name, idx, item))
    return tasks


@dataclasses.dataclass
class Component:
    """Base class for all declarative document components.

    All components are async. Use `await Document(...)` to resolve all
    async children concurrently.

    Attributes:
        condition: If Callable resolves to False, will not render the component.
    """

    condition: Callable[[], bool] = dataclasses.field(
        default=lambda: True, kw_only=True
    )

    def __await__(self) -> Generator[None, None, Self]:
        """Convenience method for awaiting a component."""
        return self.resolve().__await__()

    async def resolve(self) -> Self:
        """Recursively resolve all async children concurrently.

        Returns:
            Self with all coroutines replaced by their resolved values.
        """
        if not self.condition():
            return self

        tasks = _collect_resolve_tasks(self)

        if tasks:
            results = await asyncio.gather(*(task.awaitable for task in tasks))
            for task, result in zip(tasks, results, strict=True):
                if task.idx is None:
                    setattr(self, task.field_name, result)
                else:
                    getattr(self, task.field_name)[task.idx] = result

        return self



"""Module for handling Word comments.

Code based on sample code in https://github.com/python-openxml/python-docx/issues/93.
"""

import dataclasses
import datetime
from typing import Final
from xml.etree import ElementTree as ET

from docx import document, oxml
from docx.opc import constants as docx_constants
from docx.opc import packuri, part
from docx.oxml import ns
from docx.oxml.text import run as docx_run
from docx.text import paragraph, run
from lxml import (
    etree,  # ty:ignore[unresolved-import] # This does work; not sure why not detected.
)

_COMMENTS_PART_DEFAULT_XML_BYTES: Final[bytes] = (
    b"""
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\r
<w:comments
    xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
    xmlns:o="urn:schemas-microsoft-com:office:office"
    xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
    xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math"
    xmlns:v="urn:schemas-microsoft-com:vml"
    xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"
    xmlns:w10="urn:schemas-microsoft-com:office:word"
    xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
    xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml"
    xmlns:sl="http://schemas.openxmlformats.org/schemaLibrary/2006/main"
    xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main"
    xmlns:pic="http://schemas.openxmlformats.org/drawingml/2006/picture"
    xmlns:c="http://schemas.openxmlformats.org/drawingml/2006/chart"
    xmlns:lc="http://schemas.openxmlformats.org/drawingml/2006/lockedCanvas"
    xmlns:dgm="http://schemas.openxmlformats.org/drawingml/2006/diagram"
    xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape"
    xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup"
    xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml"
    xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml"
    xmlns:w16="http://schemas.microsoft.com/office/word/2018/wordml"
    xmlns:w16cex="http://schemas.microsoft.com/office/word/2018/wordml/cex"
    xmlns:w16cid="http://schemas.microsoft.com/office/word/2016/wordml/cid"
    xmlns:cr="http://schemas.microsoft.com/office/comments/2020/reactions">
</w:comments>
"""
).strip()


def add_comment(
    docx_doc: document.Document,
    location: tuple[paragraph.Paragraph | run.Run, paragraph.Paragraph | run.Run]
    | paragraph.Paragraph
    | run.Run,
    author: str,
    text: str,
) -> None:
    """Adds a comment to Word document.

    There is a known bug where a range of locations can be provided
    where the start comes after the end.

    Args:
        docx_doc: A Word document.
        location: The paragraph and/or run object to place the comment on.
            May also be a tuple of these where the first element is the start
            and the second element is the end.
        author: Name of the comment author.
        text: Content of the comment.

    Raises:
        ValueError: If location has incorrect typing.
    """
    if not isinstance(location, tuple):
        elements = (location._element, location._element)  # noqa: SLF001
    elif len(location) > 2:  # noqa: PLR2004
        msg = "Location must be a single element or a tuple of two."
        raise ValueError(msg)
    else:
        elements = (location[0]._element, location[1]._element)  # noqa: SLF001  # ty:ignore[unresolved-attribute]

    try:
        comments_part = docx_doc.part.part_related_by(
            docx_constants.RELATIONSHIP_TYPE.COMMENTS
        )
    except KeyError:
        # No comments part found.
        comments_part = part.Part(
            partname=packuri.PackURI("/word/comments.xml"),
            content_type=docx_constants.CONTENT_TYPE.WML_COMMENTS,
            blob=_COMMENTS_PART_DEFAULT_XML_BYTES,
            package=docx_doc.part.package,
        )
        docx_doc.part.relate_to(
            comments_part, docx_constants.RELATIONSHIP_TYPE.COMMENTS
        )

    comments_xml = oxml.parse_xml(comments_part.blob)

    # Create the comment
    comment_id = str(len(comments_xml.findall(ns.qn("w:comment"))))
    comment_element = oxml.OxmlElement("w:comment")
    comment_element.set(ns.qn("w:id"), comment_id)
    comment_element.set(ns.qn("w:author"), author)
    comment_element.set(
        ns.qn("w:date"), datetime.datetime.now(datetime.UTC).isoformat()
    )

    # Create the text element for the comment
    for para in text.split("\n"):
        comment_paragraph = oxml.OxmlElement("w:p")
        comment_run = oxml.OxmlElement("w:r")
        comment_text_element = oxml.OxmlElement("w:t")
        comment_text_element.text = para
        comment_run.append(comment_text_element)
        comment_paragraph.append(comment_run)
        comment_element.append(comment_paragraph)
        comments_xml.append(comment_element)
    comments_part._blob = ET.tostring(comments_xml)  # noqa: SLF001

    # Create the commentRangeStart and commentRangeEnd elements
    comment_range_start = oxml.OxmlElement("w:commentRangeStart")
    comment_range_start.set(ns.qn("w:id"), comment_id)
    comment_range_end = oxml.OxmlElement("w:commentRangeEnd")
    comment_range_end.set(ns.qn("w:id"), comment_id)

    # Add the commentRangeStart to the first element and commentRangeEnd to
    # the last element
    elements[0].insert(0, comment_range_start)
    elements[1].append(comment_range_end)

    # Add the comment reference to each element in the range
    # for element in elements:
    comment_reference = oxml.OxmlElement("w:r")
    comment_reference_run = oxml.OxmlElement("w:r")
    comment_reference_run_properties = oxml.OxmlElement("w:rPr")
    comment_reference_run_properties.append(
        oxml.OxmlElement("w:rStyle", {ns.qn("w:val"): "CommentReference"})
    )
    comment_reference_run.append(comment_reference_run_properties)
    comment_reference_element = oxml.OxmlElement("w:commentReference")
    comment_reference_element.set(ns.qn("w:id"), comment_id)
    comment_reference_run.append(comment_reference_element)
    comment_reference.append(comment_reference_run)

    elements[0].append(comment_reference)


@dataclasses.dataclass
class CommentRange:
    """All data representing a comment range.

    Used for temporal removal and reinsertion via `CommentPreserver`.
    """

    id: str
    start_index: int
    end_index: int


class CommentPreserver:
    """Utility to remove and reinsert comment ranges.

    As some document modifications will affect comment ranges
    in weird ways this allows to preserve them.
    """

    def __init__(self, paragraph_element: etree._Element) -> None:
        """Create CommentPreserver utility."""
        self.paragraph = paragraph_element
        self.ns = {"w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main"}

    def _get_text_length_before_element(self, element: etree._Element) -> int:
        """Calculate total text length before given element."""
        length = 0
        for elem in self.paragraph.iter():
            if elem == element:
                break
            if elem.tag == f"{{{self.ns['w']}}}t":
                length += len(elem.text or "")
        return length

    def extract_comments(self) -> list[CommentRange]:
        """Extract all comment ranges from the paragraph."""
        comments = []
        start_elements = {}

        for elem in self.paragraph.iter():
            if elem.tag == f"{{{self.ns['w']}}}commentRangeStart":
                comment_id = elem.get(f"{{{self.ns['w']}}}id")
                start_pos = self._get_text_length_before_element(elem)
                start_elements[comment_id] = start_pos

            if elem.tag == f"{{{self.ns['w']}}}commentRangeEnd":
                comment_id = elem.get(f"{{{self.ns['w']}}}id")
                if comment_id in start_elements:
                    end_pos = self._get_text_length_before_element(elem)
                    comments.append(
                        CommentRange(
                            id=comment_id,
                            start_index=start_elements[comment_id],
                            end_index=end_pos,
                        )
                    )

        return comments

    def strip_comments(self) -> None:
        """Remove all comment-related elements from the paragraph."""

        def should_remove_run(run: docx_run.CT_R) -> bool:
            """Helper to determine if a run should be removed."""
            if any(child.tag == f"{{{self.ns['w']}}}r" for child in run):
                return False
            if len(run) == 0:
                return True
            if (
                len(run) == 1
                and run[0].tag == f"{{{self.ns['w']}}}rPr"
                and any(
                    "CommentReference" in elem.get(f"{{{self.ns['w']}}}val", "")
                    for elem in run[0].iter()
                )
            ):
                return True
            has_text = any(child.tag == f"{{{self.ns['w']}}}t" for child in run)
            has_only_props = all(
                child.tag
                in [f"{{{self.ns['w']}}}rPr", f"{{{self.ns['w']}}}commentReference"]
                for child in run
            )
            return not has_text and has_only_props

        to_remove = [
            elem
            for elem in self.paragraph.iter()
            if any(tag in elem.tag for tag in ["commentRange", "commentReference"])
            or (elem.tag == f"{{{self.ns['w']}}}r" and should_remove_run(elem))
        ]

        for elem in to_remove:
            parent = elem.getparent()
            if parent is not None:
                parent.remove(elem)

        for elem in list(self.paragraph):
            if elem.tag == f"{{{self.ns['w']}}}r" and should_remove_run(elem):
                self.paragraph.remove(elem)

    @staticmethod
    def adjust_range_positions(
        comments: list[CommentRange],
        edit_start: int,
        edit_end: int,
        new_text_length: int,
    ) -> list[CommentRange]:
        """Adjust comment range positions based on text edit."""
        adjusted = []
        change_in_length = new_text_length - (edit_end - edit_start)

        for comment in comments:
            new_comment = CommentRange(
                id=comment.id,
                start_index=comment.start_index,
                end_index=comment.end_index,
            )

            # Comment completely before edit
            if comment.end_index <= edit_start:
                adjusted.append(new_comment)
                continue

            # Comment completely after edit
            if comment.start_index >= edit_end:
                new_comment.start_index += change_in_length
                new_comment.end_index += change_in_length
                adjusted.append(new_comment)
                continue

            # Comment overlaps with edit
            if comment.start_index < edit_start:
                if comment.end_index > edit_end:
                    # Edit is fully within comment range - adjust only the end position
                    new_comment.end_index = comment.end_index + change_in_length
                else:
                    # Edit starts in comment but extends beyond
                    new_comment.end_index = comment.end_index + change_in_length
            elif comment.end_index > edit_end:
                # Edit starts before comment but ends within - adjust both positions
                new_comment.start_index = edit_start
                new_comment.end_index = comment.end_index + change_in_length
            else:
                # Edit fully encompasses comment - maintain relative positions
                new_comment.start_index = edit_start
                new_comment.end_index = edit_start + new_text_length

            adjusted.append(new_comment)

        return adjusted

    def restore_comments(self, comments: list[CommentRange]) -> None:
        """Restore comment ranges to the paragraph at adjusted positions."""
        # First, ensure any leftover comment references are removed
        self.strip_comments()

        for comment in comments:
            # Create and insert comment range start
            start_elem = etree.Element(f"{{{self.ns['w']}}}commentRangeStart")
            start_elem.set(f"{{{self.ns['w']}}}id", comment.id)
            self._insert_at_position(start_elem, comment.start_index)

            # Create and insert comment range end
            end_elem = etree.Element(f"{{{self.ns['w']}}}commentRangeEnd")
            end_elem.set(f"{{{self.ns['w']}}}id", comment.id)
            self._insert_at_position(end_elem, comment.end_index)

            # Create and insert comment reference as last element
            ref_run = etree.Element(f"{{{self.ns['w']}}}r")
            ref_props = etree.SubElement(ref_run, f"{{{self.ns['w']}}}rPr")
            style = etree.SubElement(ref_props, f"{{{self.ns['w']}}}rStyle")
            style.set(f"{{{self.ns['w']}}}val", "CommentReference")
            ref = etree.SubElement(ref_run, f"{{{self.ns['w']}}}commentReference")
            ref.set(f"{{{self.ns['w']}}}id", comment.id)

            # Clean up any empty runs before adding the comment reference
            for elem in list(self.paragraph):
                if elem.tag == f"{{{self.ns['w']}}}r" and len(elem) == 0:
                    self.paragraph.remove(elem)

            # Always append reference at the end of paragraph
            self.paragraph.append(ref_run)

    def _insert_at_position(self, elem: etree._Element, text_pos: int) -> None:
        """Insert element at the specified text position."""
        runs_with_text = self._collect_text_runs()

        if not runs_with_text:
            if self.paragraph:
                self.paragraph.insert(0, elem)
            else:
                self.paragraph.append(elem)
            return

        for run_elem, start_pos, length in runs_with_text:
            if start_pos + length >= text_pos:
                if start_pos + length == text_pos:
                    insert_idx = self.paragraph.index(run_elem) + 1
                    self.paragraph.insert(insert_idx, elem)
                else:
                    insert_idx = self.paragraph.index(run_elem)
                    self.paragraph.insert(insert_idx, elem)
                return

        self.paragraph.append(elem)

    def _collect_text_runs(self) -> list[tuple]:
        """Collect all text-containing runs with their positions."""
        runs_with_text = []
        current_pos = 0
        for child in self.paragraph:
            if child.tag == f"{{{self.ns['w']}}}r":
                text_elems = child.findall(f".//{{{self.ns['w']}}}t")
                if text_elems:
                    text_length = sum(len(elem.text or "") for elem in text_elems)
                    if text_length > 0:
                        runs_with_text.append((child, current_pos, text_length))
                        current_pos += text_length
        return runs_with_text



"""Tests for the document module."""

import docx
import pytest

from cmi_docx import document, styles


def test_find_in_paragraphs() -> None:
    """Test finding a text in a document's paragraphs."""
    doc = docx.Document()
    doc.add_paragraph("Hello, world!")
    doc.add_paragraph("Hello, world, Hello!")
    extend_document = document.ExtendDocument(doc)

    actual = extend_document.find_in_paragraphs("Hello")

    assert actual[0].character_indices == [(0, 5)]
    assert actual[1].character_indices == [(0, 5), (14, 19)]


def test_find_in_runs() -> None:
    """Test finding a text in a document's runs."""
    doc = docx.Document()
    paragraph = doc.add_paragraph("Hello, world!")
    paragraph.add_run("Hello, world, Hello!")
    extend_document = document.ExtendDocument(doc)

    actual = extend_document.find_in_runs("Hello")

    assert actual[0].run_indices == (0, 0)
    assert actual[0].character_indices == (0, 5)
    assert actual[1].run_indices == (1, 1)
    assert actual[1].character_indices == (0, 5)
    assert actual[2].run_indices == (1, 1)
    assert actual[2].character_indices == (14, 19)


@pytest.mark.parametrize(
    ("runs", "needle", "replace", "expected"),
    [
        (
            ["Hello, world!"],
            "Hello",
            "Goodbye",
            "Goodbye, world!",
        ),
        (
            ["Hello, world!", " Hello!"],
            "Hello",
            "Goodbye",
            "Goodbye, world! Goodbye!",
        ),
        (
            ["Hello {{", "FULL_NAME}}"],
            "{{FULL_NAME}}",
            'Shizuka "Lea" Sakai',
            'Hello Shizuka "Lea" Sakai',
        ),
        (
            ["This is James", " Bond ", "007!"],
            "James Bond 007",
            "Patrick",
            "This is Patrick!",
        ),
        (
            ["This is Alec", " Travelyan ", "006!"],
            "This is Alec Travelyan 006!",
            "",
            "",
        ),
        (
            ["This", " is ", "Patrick!"],
            "",
            "Nonsense",
            "This is Patrick!",
        ),
    ],
)
def test_replace(runs: list[str], needle: str, replace: str, expected: str) -> None:
    """Test replacing text in a document."""
    doc = docx.Document()
    paragraph = doc.add_paragraph(runs[0])
    for run in runs[1:]:
        paragraph.add_run(run)
    extend_document = document.ExtendDocument(doc)

    extend_document.replace(needle, replace)

    assert doc.paragraphs[0].text == expected


def test_replace_with_style() -> None:
    """Test replacing text in a document with style."""
    doc = docx.Document()
    paragraph = doc.add_paragraph("")
    paragraph.add_run("{{")
    paragraph.add_run("Hello, World!")
    paragraph.add_run("}}")
    extend_document = document.ExtendDocument(doc)

    extend_document.replace("{{Hello", "Goodbye", styles.RunStyle(bold=True))

    assert doc.paragraphs[0].text == "Goodbye, World!}}"
    assert not doc.paragraphs[0].runs[0].bold
    assert doc.paragraphs[0].runs[1].bold
    assert doc.paragraphs[0].runs[1].text == "Goodbye"


def test_insert_paragraph_by_object() -> None:
    """Test inserting a paragraph into a document."""
    doc = docx.Document()
    doc.add_paragraph("Hello, world!")
    doc.add_paragraph("Goodbye, world!")
    extend_document = document.ExtendDocument(doc)
    new_paragraph = docx.Document().add_paragraph("Maintain, world!")

    extend_document.insert_paragraph_by_object(1, new_paragraph)

    assert doc.paragraphs[0].text == "Hello, world!"
    assert doc.paragraphs[1].text == "Maintain, world!"
    assert doc.paragraphs[2].text == "Goodbye, world!"


def test_insert_paragraph_by_text() -> None:
    """Test inserting a paragraph into a document."""
    doc = docx.Document()
    doc.add_paragraph("Hello, world!")
    doc.add_paragraph("Goodbye, world!")
    extend_document = document.ExtendDocument(doc)

    extend_document.insert_paragraph_by_text(1, "Maintain, world!")

    assert doc.paragraphs[0].text == "Hello, world!"
    assert doc.paragraphs[1].text == "Maintain, world!"
    assert doc.paragraphs[2].text == "Goodbye, world!"
